@@ -1,13 +1,12 @@
-use crate::enums::SignatureScheme;
+use crate::enums::{SignatureAlgorithm, SignatureScheme};
 use crate::error::Error;
 use crate::key;
-use crate::msgs::enums::SignatureAlgorithm;
 use crate::x509::{wrap_in_asn1_len, wrap_in_sequence};
 
 use ring::io::der;
+use ring::rand::{SecureRandom, SystemRandom};
 use ring::signature::{self, EcdsaKeyPair, Ed25519KeyPair, RsaKeyPair};
 
-use std::convert::TryFrom;
 use std::error::Error as StdError;
 use std::fmt;
 use std::sync::Arc;
@@ -69,57 +68,7 @@ impl CertifiedKey {
 
     /// The end-entity certificate.
     pub fn end_entity_cert(&self) -> Result<&key::Certificate, SignError> {
-        self.cert.get(0).ok_or(SignError(()))
-    }
-
-    /// Check the certificate chain for validity:
-    /// - it should be non-empty list
-    /// - the first certificate should be parsable as a x509v3,
-    /// - the first certificate should quote the given server name
-    ///   (if provided)
-    ///
-    /// These checks are not security-sensitive.  They are the
-    /// *server* attempting to detect accidental misconfiguration.
-    pub(crate) fn cross_check_end_entity_cert(
-        &self,
-        name: Option<webpki::DnsNameRef>,
-    ) -> Result<(), Error> {
-        // Always reject an empty certificate chain.
-        let end_entity_cert = self
-            .end_entity_cert()
-            .map_err(|SignError(())| {
-                Error::General("No end-entity certificate in certificate chain".to_string())
-            })?;
-
-        // Reject syntactically-invalid end-entity certificates.
-        let end_entity_cert =
-            webpki::EndEntityCert::try_from(end_entity_cert.as_ref()).map_err(|_| {
-                Error::General(
-                    "End-entity certificate in certificate \
-                                  chain is syntactically invalid"
-                        .to_string(),
-                )
-            })?;
-
-        if let Some(name) = name {
-            // If SNI was offered then the certificate must be valid for
-            // that hostname. Note that this doesn't fully validate that the
-            // certificate is valid; it only validates that the name is one
-            // that the certificate is valid for, if the certificate is
-            // valid.
-            if end_entity_cert
-                .verify_is_valid_for_dns_name(name)
-                .is_err()
-            {
-                return Err(Error::General(
-                    "The server certificate is not \
-                                             valid for the given name"
-                        .to_string(),
-                ));
-            }
-        }
-
-        Ok(())
+        self.cert.first().ok_or(SignError(()))
     }
 }
 
@@ -212,11 +161,6 @@ impl SigningKey for RsaSigningKey {
     }
 }
 
-#[allow(clippy::upper_case_acronyms)]
-#[doc(hidden)]
-#[deprecated(since = "0.20.0", note = "Use RsaSigningKey")]
-pub type RSASigningKey = RsaSigningKey;
-
 struct RsaSigner {
     key: Arc<RsaKeyPair>,
     scheme: SignatureScheme,
@@ -245,7 +189,7 @@ impl RsaSigner {
 
 impl Signer for RsaSigner {
     fn sign(&self, message: &[u8]) -> Result<Vec<u8>, Error> {
-        let mut sig = vec![0; self.key.public_modulus_len()];
+        let mut sig = vec![0; self.key.public().modulus_len()];
 
         let rng = ring::rand::SystemRandom::new();
         self.key
@@ -284,9 +228,10 @@ impl EcdsaSigningKey {
         scheme: SignatureScheme,
         sigalg: &'static signature::EcdsaSigningAlgorithm,
     ) -> Result<Self, ()> {
-        EcdsaKeyPair::from_pkcs8(sigalg, &der.0)
+        let rng = SystemRandom::new();
+        EcdsaKeyPair::from_pkcs8(sigalg, &der.0, &rng)
             .map_err(|_| ())
-            .or_else(|_| Self::convert_sec1_to_pkcs8(scheme, sigalg, &der.0))
+            .or_else(|_| Self::convert_sec1_to_pkcs8(scheme, sigalg, &der.0, &rng))
             .map(|kp| Self {
                 key: Arc::new(kp),
                 scheme,
@@ -300,6 +245,7 @@ impl EcdsaSigningKey {
         scheme: SignatureScheme,
         sigalg: &'static signature::EcdsaSigningAlgorithm,
         maybe_sec1_der: &[u8],
+        rng: &dyn SecureRandom,
     ) -> Result<EcdsaKeyPair, ()> {
         let pkcs8_prefix = match scheme {
             SignatureScheme::ECDSA_NISTP256_SHA256 => &PKCS8_PREFIX_ECDSA_NISTP256,
@@ -318,7 +264,7 @@ impl EcdsaSigningKey {
         pkcs8.extend_from_slice(&sec1_wrap);
         wrap_in_sequence(&mut pkcs8);
 
-        EcdsaKeyPair::from_pkcs8(sigalg, &pkcs8).map_err(|_| ())
+        EcdsaKeyPair::from_pkcs8(sigalg, &pkcs8, rng).map_err(|_| ())
     }
 }
 
@@ -355,7 +301,6 @@ impl SigningKey for EcdsaSigningKey {
     }
 
     fn algorithm(&self) -> SignatureAlgorithm {
-        use crate::msgs::handshake::DecomposedSignatureScheme;
         self.scheme.sign()
     }
 }
@@ -421,7 +366,6 @@ impl SigningKey for Ed25519SigningKey {
     }
 
     fn algorithm(&self) -> SignatureAlgorithm {
-        use crate::msgs::handshake::DecomposedSignatureScheme;
         self.scheme.sign()
     }
 }
